#include "jit_llvm.h"
#include "jit_llvm_p.h"

namespace GiNaC
{
using namespace llvm;
using std::string;
using std::vector;

Value* jit_context_priv::codegen(const mul& a)
{
	if (a.nops() == 0)
		throw std::invalid_argument("codegen(mul): bogus argument");

	Value* prev = codegen(a.op(0));
	for (std::size_t i = 1; i < a.nops(); i++) {
		Value* tmp = codegen(a.op(i));
		Value* curr = builder->CreateMul(prev, tmp, "multmp");
		// N.B. LLVM owns these pointers, no need to delete prev
		// (in fact, deleting it will result in error).
		prev = curr;
	}
	return prev;
}

Value* jit_context_priv::codegen(const add& a)
{
	if (a.nops() == 0)
		throw std::invalid_argument("codegen(add): bogus argument");

	llvm::Value* curr = codegen(a.op(0));

	for (std::size_t i = 1; i < a.nops(); i++) {
		llvm::Value* tmp = codegen(a.op(i));
		llvm::Value* next = builder->CreateAdd(curr, tmp, "addtmp");
		curr = next;
	}
	return curr;
}

Value* jit_context_priv::make_powi(const ex& base, const int exponent)
{
	Value* exponent_val = ConstantInt::get(Type::getInt32Ty(getGlobalContext()), exponent);
	Value* base_val = codegen(base);
	const Type* base_type = base_val->getType();
	Function* powi_fcn = 0;
	powi_fcn = Intrinsic::getDeclaration(module, Intrinsic::powi, &base_type, 1);
	if (!powi_fcn)
		throw std::runtime_error("Could not get found powi function");

	vector<Value*> args;
	args.push_back(base_val);
	args.push_back(exponent_val);

	Value* op = builder->CreateCall(powi_fcn, args.begin(), args.end(), "tmp_powi");
	return op;
}

Value* jit_context_priv::make_pow(const ex& base, const ex& exponent)
{
	Value* exponent_val = codegen(exponent);
	Value* base_val = codegen(base);
	const Type* base_type = base_val->getType();
	Function* pow_fcn = 0;
	pow_fcn = Intrinsic::getDeclaration(module, Intrinsic::pow, &base_type, 1);
	if (!pow_fcn)
		throw std::runtime_error("Could not get found pow function");

	vector<Value*> args;
	args.push_back(base_val);
	args.push_back(exponent_val);

	Value* op = builder->CreateCall(pow_fcn, args.begin(), args.end(), "tmp_pow");
	return op;
}

Value* jit_context_priv::codegen(const power& p)
{
	if (p.op(1).info(info_flags::integer)) {
		const int exponent = ex_to<numeric>(p.op(1)).to_int();
		return make_powi(p.op(0), exponent);
	} else {
		return make_pow(p.op(0), p.op(1));
	}
}

// TODO: check if the GiNaC function has C/C++ counterpart
Value* jit_context_priv::codegen(const function& gf)
{
	string fname = gf.get_name();
	vector<Value*> args(gf.nops());

	for (std::size_t i = 0; i < gf.nops(); ++i)
		args[i] = codegen(gf.op(i));

	vector<const Type*> args_decl(gf.nops(), Type::getDoubleTy(getGlobalContext()));
	FunctionType* fproto =
		FunctionType::get(Type::getDoubleTy(getGlobalContext()), args_decl, false);
	Function* f = static_cast<Function*>(module->getOrInsertFunction(fname, fproto));
	Value* ret = builder->CreateCall(f, args.begin(), args.end(), "tmp_arg");
	return ret;
}

/**
 * Symbols are assumed to be function arguments.
 */
Value* jit_context_priv::codegen(const symbol& s)
{
	exvalmap::const_iterator i = cache.find(s);
	if (i != cache.end())
		return i->second;

	// foo(double* args, int n_args)
	if (pass_as_array) {
		std::size_t i_ = 0;
		while (!vars[i_].is_equal(s))
			++i_;
		if (i_ >= vars.size())
			throw std::logic_error(string("unknown symbol \"") + s.get_name() + "\"");

		Value* i = ConstantInt::get(Type::getInt32Ty(getGlobalContext()), i_);
		Function::arg_iterator ai = fp->arg_begin();
		Value* ops = ai;
		Value* ptr = builder->CreateGEP(ops, i, "arg_ptr");
		Value* val = builder->CreateLoad(ptr, "arg_val");

		cache[s] = val;
		return val;
	}

	Function::arg_iterator ai = fp->arg_begin();
	for (std::size_t i = 0; i != vars.size(); ++ai, ++i) {
		if (vars[i].is_equal(s)) {
			cache[s] = ai;
			return ai;
		}
	}

	throw std::logic_error(string("unknown symbol \"") + s.get_name() + "\"");
}

Value* jit_context_priv::codegen(const ex& e)
{
	if (is_a<add>(e))
		return codegen(ex_to<add>(e));
	else if (is_a<mul>(e))
		return codegen(ex_to<mul>(e));
	else if (is_a<numeric>(e)) {
		const numeric n = ex_to<numeric>(e);
		return ConstantFP::get(Type::getDoubleTy(getGlobalContext()), n.to_double());
	} else if (is_a<symbol>(e))
		return codegen(ex_to<symbol>(e));
	else if (is_a<power>(e))
		return codegen(ex_to<power>(e));
	else if (is_a<function>(e))
		return codegen(ex_to<function>(e));

	string class_name(ex_to<basic>(e).class_name());
	throw std::logic_error(class_name + string(": JIT compilation is not implemented"));
}

/**
 * Make a function prototype for passing arguments via pointer.
 *
 * double name(double* args, int32_t n_args)
 *
 * @param name name of newly created function
 * @param n_args number of arguments
 */
Function* jit_context_priv::make_proto_ptr(const string& name, const unsigned n_args)
{
	std::vector<const Type*> arg_types;
	// double* args
	arg_types.push_back(PointerType::getUnqual(Type::getDoubleTy(getGlobalContext())));
	// int32_t n_args
	arg_types.push_back(Type::getInt32Ty(getGlobalContext()));
	FunctionType* ft = FunctionType::get(Type::getDoubleTy(getGlobalContext()), arg_types, false);
	Function* f = Function::Create(ft, 
		                       Function::ExternalLinkage, name,
		                       module);
	// Make sure only one function with given name and number of args exists
	ensure_odr(name, n_args, f);

	// Give arguments human readable names.
	Function::arg_iterator ai = f->arg_begin();
	ai->setName("args");
	++ai;
	ai->setName("n_args");
	return f;
}

/**
 * Make a function prototype for passing arguments by values.
 *
 * double name(double arg1, double arg2, double arg3);
 *
 * @param name name of newly created function
 * @param vars arguments' names
 * @param ctx context (AKA module)
 */
Function* jit_context_priv::make_proto_val(const string& name, const vector<string>& vars)
{

	std::vector<const Type*> arg_types(vars.size(), Type::getDoubleTy(getGlobalContext()));
	FunctionType* ft = FunctionType::get(Type::getDoubleTy(getGlobalContext()),
					     arg_types, false);
	Function* f = Function::Create(ft, Function::ExternalLinkage,
				       name, module);
	Function::arg_iterator ai = f->arg_begin();
	for (std::size_t i = 0; i != vars.size(); ++ai, ++i)
		ai->setName(vars[i]);
	return f;
}


void jit_context_priv::ensure_odr(const string& name, const unsigned n_args, Function* f)
{
	if (f->getName() != name) {
		f->eraseFromParent();
		f = module->getFunction(name);

		if (!f->empty()) 
			throw std::logic_error("redefinition of function");

		if (f->arg_size() != n_args)
			throw std::logic_error("redefinition of function with different # args");
	}
}

/**
 * Compile expression e into LLVM function.
 *
 * @param e expression to operate on.
 * @param name name of LLVM function which will be produced.
 * @param vars arguments of function (each should be a symbol).
 * @param safe validate the result. Defaults to off, because verification
 *        is way too slow.
 * @return LLVM function corresponding to the expression.
 */
void jit_context_priv::operator()(const ex& e,
				  const string& name,
				  const exvector& vars_,
				  const bool safe)
{
	cache.clear();
	vars = vars_;
	if (vars_.size() > 3) {
		pass_as_array = true;
		fp = make_proto_ptr(name, vars.size());
	} else {
		pass_as_array = false;
		std::vector<std::string> args_names(vars_.size());
		for (std::size_t i = 0; i < vars_.size(); i++)
			args_names[i] = GiNaC::ex_to<GiNaC::symbol>(vars_[i]).get_name();

		fp = make_proto_val(name, args_names);
	}

	if (!fp)
		throw std::runtime_error("Failed to create the prototype");

	BasicBlock* bb = BasicBlock::Create(getGlobalContext(), "entry", fp);
	builder->SetInsertPoint(bb);

	Value* retval = 0;
	retval = codegen(e);
	if (!retval)
		throw std::runtime_error("failed to process the function body");
	builder->CreateRet(retval);
	if (safe)
		verifyFunction(*fp);
	pass_manager->run(*fp);
}

template<typename T> static string toString(const T& what)
{
	std::ostringstream os;
	os << what;
	return os.str();
}

void jit_context_priv::save(const string& file_name)
{
	if (!fp)
		throw std::runtime_error("The module has no expressions so far");

	string errStr;
	llvm::raw_fd_ostream file(file_name.c_str(), errStr);
	bool ok = true;
	try {
		module->print(file, 0);
		file.flush();
		file.close();
	} catch (...) {
		file.close();
		ok = false;
	}
	if (!ok)
		throw std::runtime_error("failed to save the module");
}

void* jit_context_priv::get_native()
{
	std::string fname = std::string("GiNaC_jit") + toString(compile_count) + ".ll";
	try {
		save(fname);
	} catch (...) { 
		// ignore
	}
	++compile_count;

	cache.clear();
	return engine->getPointerToFunction(fp);
}

jit_context_priv::jit_context_priv()
{
	// XXX: how do I set register allocator properly?
	char* reg_alloc_argv[] = { "dummy", "-regalloc=local" };
	int reg_alloc_argc = sizeof(reg_alloc_argv)/sizeof(char*);
	const char* progname = "GiNaC jit";
	cl::ParseCommandLineOptions(reg_alloc_argc, reg_alloc_argv, progname);
		

	compile_count = 0;
	InitializeNativeTarget();
	module = new Module("GiNaC_jit", getGlobalContext());
	if (!module)
		throw std::runtime_error("failed to create the module");
	builder = new IRBuilder<>(getGlobalContext());
	if (!builder) {
		delete module;
		throw std::runtime_error("failed to create IR builder");
	}
	std::string errStr;
	engine = EngineBuilder(module).setOptLevel(CodeGenOpt::None).setErrorStr(&errStr).create();
	if (!engine) {
		delete builder;
		delete module;
		throw std::runtime_error("failed to create execution engine: " + errStr);
	}
	pass_manager = new FunctionPassManager(module);
	if (!pass_manager) {
		delete builder;
		delete module;
		// N.B. engine is owned by module, so it gets deleted too
	}

	pass_manager->add(createGVNPass());
	pass_manager->add(createInstructionCombiningPass());
	pass_manager->doInitialization();
	RegisterRegAlloc::setDefault(&createLocalRegisterAllocator);
	fp = 0;
	vars = exvector();
	pass_as_array = false;
	cache = std::map<ex, Value*, ex_is_less>();
}

jit_context_priv::~jit_context_priv()
{
	// N.B. no delete engine; -- it's owned by module
	delete builder;
	delete module;
}

jit_context::jit_context()
{
	impl = new jit_context_priv;
}

jit_context::~jit_context()
{
	delete impl;
}

void jit_context::operator()(const ex& e,
			     const string& fname,
			     const exvector& vars,
			     const bool safe)
{
	(*impl)(e, fname, vars, safe);
}

void jit_context::save(const string& file_name)
{
	impl->save(file_name);
}

void* jit_context::get_native()
{
	return impl->get_native();
}

} // namespace GiNaC

