#ifndef GINAC_JIT_LLVM_PRIVATE_H
#define GINAC_JIT_LLVM_PRIVATE_H
#include <cstddef>
#include <stdexcept>
#include "jit_llvm.h"
#include "ginac/symbol.h"
#include "ginac/add.h"
#include "ginac/mul.h"
#include "ginac/power.h"
#include "ginac/function.h"
#include <sstream>
#include <fstream>
#include <llvm/DerivedTypes.h>
#include <llvm/ExecutionEngine/ExecutionEngine.h>
#include <llvm/ExecutionEngine/JIT.h>
#include <llvm/Target/TargetData.h>
#include <llvm/Target/TargetSelect.h>
#include <llvm/LLVMContext.h>
#include <llvm/PassManager.h>
#include <llvm/Transforms/Scalar.h>
#include <llvm/Module.h>
#include <llvm/Analysis/Verifier.h>
#include <llvm/Support/IRBuilder.h>
#include <llvm/Support/raw_ostream.h>
#include <llvm/Intrinsics.h>
#include <llvm/CodeGen/RegAllocRegistry.h>
#include <llvm/CodeGen/Passes.h>
#include <llvm/Support/CommandLine.h>

namespace GiNaC
{
using namespace llvm;
using std::string;
using std::vector;

typedef std::map<ex, Value*, ex_is_less> exvalmap;

struct jit_context_priv
{
	Module* module;
	IRBuilder<> * builder;
	ExecutionEngine* engine;
	FunctionPassManager* pass_manager;
	Function* fp;
	exvector vars;
	bool pass_as_array;
	exvalmap cache;
	int compile_count;
	
	jit_context_priv();
	~jit_context_priv();

	/**
	 * @brief Compile GiNaC expression into LLVM IR, apply basic optimizations
	 *
	 * @param e expression to compile
	 * @param fname name of native function to create
	 * @param vars_ symbols denoting function arguments
	 * @param safe Ask LLVM to run basic sanity checks on function.
	 *        Switched off by default since these checks are way too
	 *        expensive for any real-life expression.
	 *
	 * Substituting floating-point values for the arguments (@a vars_)
	 * in the expression @a e must yeld a real number. Any functions
	 * referenced by @e must be either
	 * - present in the C runtime library or other libraries linked
	 *   into the main binary
	 * - previously compiled in the same jit_context
	 */
	void operator()(const ex& e,
			const string& name,
			const exvector& vars_,
			const bool safe);
	/**
	 * @brief Compile previously converted expression into a native code
	 */
	void* get_native();

	/**
	 * @brief Save previously converted expression in a form of LLVM asm 
	 */
	void save(const string& file_name);

	/**
	 * @brief Convert expression into LLVM IR
	 *
	 * As of now it handles sums, products, powers, and functions.
	 * (GiNaC) functions are naively replaced with external function calls,
	 * therefore C/C++ runtime library should provide them.
	 */
	Value* codegen(const ex& e);
	/**
	 * @brief Convert a sum into LLVM IR
	 */
	Value* codegen(const add& a);
	/**
	 * @brief Convert a product into LLVM IR
	 */
	Value* codegen(const mul& m);
	/**
	 * @brief Convert a power into LLVM IR
	 */
	Value* codegen(const power& p);
	/**
	 * @brief Convert a power with integer exponent into LLVM IR
	 */
	Value* make_powi(const ex& base, const int exponent);
	/**
	 * @brief Convert a power with non-integer exponent into LLVM IR
	 */
	Value* make_pow(const ex& base, const ex& exponent);
	/**
	 * @brief Convert a (GiNaC::) function into LLVM IR
	 *
	 * Calls an external function with corresponding arguments.
	 * The external function must be provided either by C/C++ runtime
	 * or your binary. Note: there's no check if such external function
	 * exists, be careful.
	 */
	Value* codegen(const function& f);

	/**
	 * @brief Convert a symbol into LLVM IR
	 *
	 * @note Symbols are assumed to be function arguments.
	 */
	Value* codegen(const symbol& s);

	/**
	 * Ensure one definition rule (ODR) holds.
	 *
	 * Make sure only one function with given name and number of args exists
	 * withing the current module.
	 */
	void ensure_odr(const string& name, const unsigned n_args, Function* f);
	/**
	 * Make a function prototype for passing arguments via pointer.
	 *
	 * double name(double* args, int32_t n_args)
	 *
	 * @param name name of newly created function
	 * @param n_args number of arguments
	 * @param ctx context (AKA module)
	 */
	Function* make_proto_ptr(const string& name, const unsigned n_args);
	/**
	 * Make a function prototype for passing arguments by values.
	 *
	 * double name(double arg1, double arg2, double arg3);
	 *
	 * @param name name of newly created function
	 * @param vars arguments' names
	 */
	Function* make_proto_val(const string& name, const vector<string>& vars);
};

} // namespace GiNaC

#endif /* GINAC_JIT_LLVM_PRIVATE_H */

